==> cui-spa-app-class-0.1.5.tgz: package/modules/dataInterface.js <==
class DATA_INTERFACE_DATA_SOURCE {
    constructor(oConfig, fOnlineStatus){
        this.oConfig = oConfig;
        this.fOnlineStatus = fOnlineStatus;
    }   
}

class DATA_INTERFACE_SERVICE_EXTERNAL extends DATA_INTERFACE_DATA_SOURCE{

    constructor(oConfig, fOnlineStatus){
        super(oConfig, fOnlineStatus)
        
        this.oDefaultParams = {
            method:"POST",
            headers: {
                'Content-Type': 'application/json',                
            },            
        }        

        this.dataSourceType = "service";

        this.oInitConfig = (this.oConfig) ? this.oConfig : {};
        this.sEndpoint = (this.oInitConfig.url)? this.oInitConfig.url:"";
        this.oInitConfig.params = (this.oInitConfig.params) ? this.oInitConfig.params : {};
        this.oData = (this.oInitConfig.params.data) ? this.oInitConfig.params.data : {};

        this.oParams = Object.assign(this.oDefaultParams);
        this.oParams.method = (this.oInitConfig.params.method) ? this.oInitConfig.params.method : this.oParams.method;        
    }

    async request(oRequestConfig, bFlushBody = false){
        if (this.fOnlineStatus()) {
            return this.request_online(oRequestConfig, bFlushBody);
        } else {
            return this.request_offline(oRequestConfig, bFlushBody);
        }
    }

    async request_online(oRequestConfig, bFlushBody){
        let sRequestEndpoint = this.sEndpoint;
        let oRequestParams = Object.assign(this.oParams);
        let oRequestData = this.oData;

        if(oRequestConfig){
            if(oRequestConfig.method){
                oRequestParams.method = oRequestConfig.method;
            }
            if(oRequestConfig.data){
                if(bFlushBody){
                    oRequestData = oRequestConfig.data;
                }
                else{
                    oRequestData = Object.assign(oRequestData, oRequestConfig.data);
                }
            }
        }  
        
        if(oRequestParams.method == "GET"){
            let uQueryParams = new URLSearchParams(oRequestData);
            let sQueryString = uQueryParams.toString();
            if(sQueryString.length > 0){
                sRequestEndpoint += "?" + sQueryString;    
            }
        }
        else if(oRequestParams.method == "POST"){
            oRequestParams.body = JSON.stringify(oRequestData);
        }
    
        return fetch(sRequestEndpoint, oRequestParams)  
        .then( response => response.json())      
        .catch( error => { throw new Error(error) });
    } 

    async request_offline(){
        /*DEBUG: START*/
        console.log("DATA_INTERFACE_SERVICE_EXTERNAL - request_offline()");
        /*DEBUG: END*/   
    }    

}

class DATA_INTERFACE_SERVICE_LOCAL extends DATA_INTERFACE_DATA_SOURCE{
    constructor(oConfig, fOnlineStatus){
        super(oConfig, fOnlineStatus);
        this.cStore = oConfig.store;
        this.dataSourceType = "dataStore";
    } 

    setState(oNewState){

        return this.cStore.setState(oNewState);
    }

    getState(){
        return this.cStore.getState();
    }


    setValue(sDataKey, sNewValue){

        return this.cStore.setValue(sDataKey, sNewValue);
    }

    getValue(sDataKey){
        /*DEBUG: START*/
        console.log("DATA_INTERFACE_SERVICE_LOCAL - getValue", sDataKey);
        /*DEBUG: END*/
        return this.cStore.getValue(sDataKey);
    }

}

export default class DATA_INTERFACE {
/*
    Responsibilities:
        1. Register sources for the page. This includes services and local store values. Sources will be registered with baseline params. Per individial request they can be overritten. 
        2. Populate ScreenSchema with data to bind. 
        3. Request single pieces of data from a registered source. 

        4. Sending update requests to a registered source. 
        5. Subscribing to update events at registered source (Store/Web socket);


    //SourceMap: - May define to own class. 
        Object containing a source name and data keys for defining a value. 
        {
            "sSourceName":null,
            "aDataPath":null
        }

    //ServiceMap:
        {
            "sServiceName":"",
            "cService":serviceOBJ,
            "aDataPath":["owners", "primary", "fName"]
        }

*/
    constructor() {
        this.oSources = {}; // Will be depreciated!, Shifting to dataSources
        this.oOnline = {
            "status": true
        }; //Working in online or offline mode. Will determine where each source resolves. 

        this.oDataSources = {};

    }

    init() {
        console.log("DATA_INTERFACE INIT");
    }

    getOnlineStatus(){
        return this.oOnline.status;
    }

    // ======================
    // Data Source methods
    // ======================
    _registerDataSource(sSourceName, oProps, bOverride = false, sourceType){

        /*DEBUG: START*/
        console.log('DATA_INTERFACE - registerService:', sSourceName, oProps, bOverride);
        /*DEBUG: END*/

        if(!sSourceName || !oProps){
            return false;
        } 

        if(!this.oDataSources.hasOwnProperty(sSourceName) || this.oDataSources.hasOwnProperty(sSourceName) && bOverride){

            let cNewService;
            if(sourceType == "service"){
                cNewService = new DATA_INTERFACE_SERVICE_EXTERNAL(oProps, this.getOnlineStatus.bind(this));  
            }
            else if(sourceType == "dataStore"){
                cNewService = new DATA_INTERFACE_SERVICE_LOCAL(oProps, this.getOnlineStatus.bind(this));  
            }

            if(cNewService){
                this.oDataSources[sSourceName] = cNewService;
                return cNewService; 
            }           
        }
    }

    registerService(sServiceName, oProps, bOverride = false){
        return this._registerDataSource(sServiceName, oProps, bOverride, 'service');
    }

    registerDataStore(sStoreName, oProps, bOverride = false){
        return this._registerDataSource(sStoreName, oProps, bOverride, 'dataStore');
    }

    getSource(sSourceName){
        if(this.oDataSources[sSourceName]){
            return this.oDataSources[sSourceName];    
        }

        return null; 
    }

    getSources(){
        return this.oDataSources;
    }

    //May want to copy sources in the future??
    getCopyOfSource(sSourceName){
        return // New service populated with config from existing service. 
    }

    registerSource(sSourceName, cDataSource){
        if(this.oDataSources[sSourceName]){
            this.oDataSources[sSourceName] = cDataSource;    
        }
    }
    
    // ======================
    // Full Schema processing
    // ======================

    //Processes given schema resolving against any registered sources. Data results are added into the supplied schema and returned. 
    async populateSchema(oScreenSchema){
        let oKeysPerSource = {};

        let pDataRequests = [];

        //Parse screen schema splitting viewkeys by source. 
        if(oScreenSchema){
            for (const [viewKey, data] of Object.entries(oScreenSchema)) {
                if(data.source){
                    let oSourceMap = this._createSourceMap(data.source);  
                    if(oSourceMap){
                        oScreenSchema[viewKey].oSourceMap = oSourceMap;

                        //Split into sources array for processing. 
                        if(oSourceMap.sSourceName){

                            if(!oKeysPerSource[oSourceMap.sSourceName]){
                                oKeysPerSource[oSourceMap.sSourceName] = [];
                            }
                            oKeysPerSource[oSourceMap.sSourceName].push(data);
                        }                        
                    }                     
                }                
            }
            
            //Process oKeysPerSource here. 
            // Split to own function 
            if(Object.keys(oKeysPerSource).length>0){
                for (const [sDataSourceName, aDataToRetreive] of Object.entries(oKeysPerSource)) {

                    //This should be generating promises and queuing them to a resolve at the end of the function. 

                    let oSource = this.getSource(sDataSourceName);

                    if(oSource){      
                        if(oSource.dataSourceType =="dataStore"){                            
                            // Split to own function 
                            // Datastore lookups are not batched, they are retreived per field. 
                            for(let i=0;i<aDataToRetreive.length; i++){
                                let oField = aDataToRetreive[i];

                                let sourceRetreivedValue = oSource.cStore.getValue(oField.oSourceMap.aDataPath);

                                if(sourceRetreivedValue){
                                    oField.value = sourceRetreivedValue;    
                                }                                
                            }
                        }
                        else if(oSource.dataSourceType == "service"){
                            /*DEBUG: START*/
                            console.log('Retreive data from service for:', aDataToRetreive);
                            /*DEBUG: END*/
                            
                            let responseData = oSource.request().then((response)=>{
                                this.mapDataToKeys(aDataToRetreive, response);
                            });
                            //Parse respose data with fields. 
                            pDataRequests.push(responseData);
                        }
                    }
                }  
            }
        }      

        // If there are any promist requests pending, wait for results before proceeding. 
        if(pDataRequests.length>0){
            await Promise.all(pDataRequests).then((values) => {
              console.log('populateSchema - promise all:', values);
            });    
        }

        return oScreenSchema;
    }

    mapDataToKeys(keys, response){
        /*DEBUG: START*/
        console.log("mapDataToKeys - Need to refine. Use store nestedValue function. ", keys, response);
        /*DEBUG: END*/

        for(let i=0; i<keys.length;i++){
            keys[i].value = this._checkNestedValue(response.result[0].body,keys[i].oSourceMap.aDataPath);
        }
    }

    //TODO:
    //Should be abstracted to utils library as it is also used in dataStore
    _checkNestedValue(currentObj, dataKey){
        let nestedValue = null;
        let nestedDataKey = [...dataKey];//Shalow copy since we don't want to destruct original key array. 
        
        if(nestedDataKey.length > 1){
            let currentKey = nestedDataKey.shift();
            if(currentObj.hasOwnProperty(currentKey)){
                if(nestedDataKey.length>0){
                    nestedValue = this._checkNestedValue(currentObj[currentKey], nestedDataKey);
                }    
                else{
                    return currentObj[currentKey];
                }
            }
            else{
                return null;            
            }                       
        }
        else{
            if(currentObj.hasOwnProperty(nestedDataKey)){
                return currentObj[nestedDataKey];    
            }
            else{
                return null;
            }
        }

        return nestedValue;
    }

    // Returns an object with sSourceName and an aDataPath array containing location keys within the source. 
    _createSourceMap(sSource){
        let oSourceMap = {
            "sSourceName":null,
            "aDataPath":null
        }

        if(sSource.indexOf('.')>0){
            let aSourceKeys = sSource.split('.');
            
            oSourceMap.sSourceName = aSourceKeys[0];
            aSourceKeys.shift();//Remove source and save remaining keys to determine data key at source

            if(aSourceKeys.length >= 0){
                oSourceMap.aDataPath = aSourceKeys;
            }            
        }
        else{
            oSourceMap.sSourceName = sSource;
        }

        return oSourceMap;        
    }

    //Used to retreive a single piece of data from a source. 

    //sFullSource resolves to a piece of data/object. sDataSet can be used to retreive multiple properties from that data/object.

    /*
        sFullSource = "propertyInfo.address"
        aDataSet= {
            "city": "city",
            "state": "state",    
            "street": "street1"
            "ownerName":"structure1.color"
        };

        would return a populated data response composed of:
        {
            "city": "propertyInfo.address.city",
            "state": "propertyInfo.address.state",    
            "street": "propertyInfo.address.street1"
            "ownerName":"propertyInfo.address.structure1.color"
        };

    */   

}

==> cui-spa-app-class-0.2.3.tgz: package/modules/dataInterface.js <==
class DATA_INTERFACE_DATA_SOURCE {
    constructor(oConfig, fOnlineStatus){
        this.oConfig = oConfig;
        this.fOnlineStatus = fOnlineStatus;
    }   
}

class DATA_INTERFACE_SERVICE_EXTERNAL extends DATA_INTERFACE_DATA_SOURCE{

    constructor(oConfig, fOnlineStatus){
        super(oConfig, fOnlineStatus)
        
        this.oDefaultParams = {
            method:"POST",
            headers: {
                'Content-Type': 'application/json',                
            },
            cache: "no-cache"      
        }        

        this.dataSourceType = "service";

        this.oInitConfig = (this.oConfig) ? this.oConfig : {};
        this.sEndpoint = (this.oInitConfig.url)? this.oInitConfig.url:"";
        this.oInitConfig.params = (this.oInitConfig.params) ? this.oInitConfig.params : {};
        this.oData = (this.oInitConfig.params.data) ? this.oInitConfig.params.data : {};

        this.oParams = Object.assign(this.oDefaultParams);
        this.oParams.method = (this.oInitConfig.params.method) ? this.oInitConfig.params.method : this.oParams.method;        
    }

    async request(oRequestConfig, bFlushBody = false){
        if (this.fOnlineStatus()) {
            return this.request_online(oRequestConfig, bFlushBody);
        } else {
            return this.request_offline(oRequestConfig, bFlushBody);
        }
    }

    async request_online(oRequestConfig, bFlushBody){
        let sRequestEndpoint = this.sEndpoint;
        let oRequestParams = Object.assign(this.oParams);
        let oRequestData = this.oData;

        if(oRequestConfig){
            if(oRequestConfig.method){
                oRequestParams.method = oRequestConfig.method;
            }
            if(oRequestConfig.data){
                if(bFlushBody){
                    oRequestData = oRequestConfig.data;
                }
                else{
                    oRequestData = Object.assign(oRequestData, oRequestConfig.data);
                }
            }
        }  
        
        if(oRequestParams.method == "GET"){
            let uQueryParams = new URLSearchParams(oRequestData);
            let sQueryString = uQueryParams.toString();
            if(sQueryString.length > 0){
                sRequestEndpoint += "?" + sQueryString;    
            }
        }
        else if(oRequestParams.method == "POST"){
            oRequestParams.body = JSON.stringify(oRequestData);
        }
    
        return fetch(sRequestEndpoint, oRequestParams)  
        .then( response => response.json())      
        .catch( error => { throw new Error(error) });
    } 

    async request_offline(){
        /*DEBUG: START*/
        console.log("DATA_INTERFACE_SERVICE_EXTERNAL - request_offline()");
        /*DEBUG: END*/   
    }    

}

class DATA_INTERFACE_SERVICE_LOCAL extends DATA_INTERFACE_DATA_SOURCE{
    constructor(oConfig, fOnlineStatus){
        super(oConfig, fOnlineStatus);
        this.cStore = oConfig.store;
        this.dataSourceType = "dataStore";
    } 

    setState(oNewState){

        return this.cStore.setState(oNewState);
    }

    getState(){
        return this.cStore.getState();
    }


    setValue(sDataKey, sNewValue){

        return this.cStore.setValue(sDataKey, sNewValue);
    }

    getValue(sDataKey){
        /*DEBUG: START*/
        console.log("DATA_INTERFACE_SERVICE_LOCAL - getValue", sDataKey);
        /*DEBUG: END*/
        return this.cStore.getValue(sDataKey);
    }

}

export default class DATA_INTERFACE {
/*
    Responsibilities:
        1. Register sources for the page. This includes services and local store values. Sources will be registered with baseline params. Per individial request they can be overritten. 
        2. Populate ScreenSchema with data to bind. 
        3. Request single pieces of data from a registered source. 

        4. Sending update requests to a registered source. 
        5. Subscribing to update events at registered source (Store/Web socket);


    //SourceMap: - May define to own class. 
        Object containing a source name and data keys for defining a value. 
        {
            "sSourceName":null,
            "aDataPath":null
        }

    //ServiceMap:
        {
            "sServiceName":"",
            "cService":serviceOBJ,
            "aDataPath":["owners", "primary", "fName"]
        }

*/
    constructor() {
        this.oSources = {}; // Will be depreciated!, Shifting to dataSources
        this.oOnline = {
            "status": true
        }; //Working in online or offline mode. Will determine where each source resolves. 

        this.oDataSources = {};

    }

    init() {
        console.log("DATA_INTERFACE INIT");
    }

    getOnlineStatus(){
        return this.oOnline.status;
    }

    // ======================
    // Data Source methods
    // ======================
    _registerDataSource(sSourceName, oProps, bOverride = false, sourceType){

        /*DEBUG: START*/
        console.log('DATA_INTERFACE - registerService:', sSourceName, oProps, bOverride);
        /*DEBUG: END*/

        if(!sSourceName || !oProps){
            return false;
        } 

        if(!this.oDataSources.hasOwnProperty(sSourceName) || this.oDataSources.hasOwnProperty(sSourceName) && bOverride){

            let cNewService;
            if(sourceType == "service"){
                cNewService = new DATA_INTERFACE_SERVICE_EXTERNAL(oProps, this.getOnlineStatus.bind(this));  
            }
            else if(sourceType == "dataStore"){
                cNewService = new DATA_INTERFACE_SERVICE_LOCAL(oProps, this.getOnlineStatus.bind(this));  
            }

            if(cNewService){
                this.oDataSources[sSourceName] = cNewService;
                return cNewService; 
            }           
        }
    }

    registerService(sServiceName, oProps, bOverride = false){
        return this._registerDataSource(sServiceName, oProps, bOverride, 'service');
    }

    registerDataStore(sStoreName, oProps, bOverride = false){
        return this._registerDataSource(sStoreName, oProps, bOverride, 'dataStore');
    }

    getSource(sSourceName){
        if(this.oDataSources[sSourceName]){
            return this.oDataSources[sSourceName];    
        }

        return null; 
    }

    getSources(){
        return this.oDataSources;
    }

    //May want to copy sources in the future??
    getCopyOfSource(sSourceName){
        return // New service populated with config from existing service. 
    }

    registerSource(sSourceName, cDataSource){
        if(this.oDataSources[sSourceName]){
            this.oDataSources[sSourceName] = cDataSource;    
        }
    }
    
    // ======================
    // Full Schema processing
    // ======================

    //Processes given schema resolving against any registered sources. Data results are added into the supplied schema and returned. 
    async populateSchema(oScreenSchema){
        let oKeysPerSource = {};

        let pDataRequests = [];

        //Parse screen schema splitting viewkeys by source. 
        if(oScreenSchema){
            for (const [viewKey, data] of Object.entries(oScreenSchema)) {
                if(data.source){
                    let oSourceMap = this._createSourceMap(data.source);  
                    if(oSourceMap){
                        oScreenSchema[viewKey].oSourceMap = oSourceMap;

                        //Split into sources array for processing. 
                        if(oSourceMap.sSourceName){

                            if(!oKeysPerSource[oSourceMap.sSourceName]){
                                oKeysPerSource[oSourceMap.sSourceName] = [];
                            }
                            oKeysPerSource[oSourceMap.sSourceName].push(data);
                        }                        
                    }                     
                }                
            }
            
            //Process oKeysPerSource here. 
            // Split to own function 
            if(Object.keys(oKeysPerSource).length>0){
                for (const [sDataSourceName, aDataToRetreive] of Object.entries(oKeysPerSource)) {

                    //This should be generating promises and queuing them to a resolve at the end of the function. 

                    let oSource = this.getSource(sDataSourceName);

                    if(oSource){      
                        if(oSource.dataSourceType =="dataStore"){                            
                            // Split to own function 
                            // Datastore lookups are not batched, they are retreived per field. 
                            for(let i=0;i<aDataToRetreive.length; i++){
                                let oField = aDataToRetreive[i];

                                let sourceRetreivedValue = oSource.cStore.getValue(oField.oSourceMap.aDataPath);

                                if(sourceRetreivedValue){
                                    oField.value = sourceRetreivedValue;    
                                }                                
                            }
                        }
                        else if(oSource.dataSourceType == "service"){
                            /*DEBUG: START*/
                            console.log('Retreive data from service for:', aDataToRetreive);
                            /*DEBUG: END*/
                            
                            let responseData = oSource.request().then((response)=>{
                                this.mapDataToKeys(aDataToRetreive, response);
                            });
                            //Parse respose data with fields. 
                            pDataRequests.push(responseData);
                        }
                    }
                }  
            }
        }      

        // If there are any promist requests pending, wait for results before proceeding. 
        if(pDataRequests.length>0){
            await Promise.all(pDataRequests).then((values) => {
              console.log('populateSchema - promise all:', values);
            });    
        }

        return oScreenSchema;
    }

    mapDataToKeys(keys, response){
        /*DEBUG: START*/
        console.log("mapDataToKeys - Need to refine. Use store nestedValue function. ", keys, response);
        /*DEBUG: END*/

        for(let i=0; i<keys.length;i++){

            let responseData = this._checkNestedValue(response.result[0].body,keys[i].oSourceMap.aDataPath);

            for (const [viewKeyProp, viewPropValue] of Object.entries(responseData)) {
               keys[i][viewKeyProp] = viewPropValue;
            }
        }
    }

    //TODO:
    //Should be abstracted to utils library as it is also used in dataStore
    _checkNestedValue(currentObj, dataKey){
        let nestedValue = null;
        let nestedDataKey = [...dataKey];//Shalow copy since we don't want to destruct original key array. 
        
        if(nestedDataKey.length > 1){
            let currentKey = nestedDataKey.shift();
            if(currentObj.hasOwnProperty(currentKey)){
                if(nestedDataKey.length>0){
                    nestedValue = this._checkNestedValue(currentObj[currentKey], nestedDataKey);
                }    
                else{
                    return currentObj[currentKey];
                }
            }
            else{
                return null;            
            }                       
        }
        else{
            if(currentObj.hasOwnProperty(nestedDataKey)){
                return currentObj[nestedDataKey];    
            }
            else{
                return null;
            }
        }

        return nestedValue;
    }

    // Returns an object with sSourceName and an aDataPath array containing location keys within the source. 
    _createSourceMap(sSource){
        let oSourceMap = {
            "sSourceName":null,
            "aDataPath":null
        }

        if(sSource.indexOf('.')>0){
            let aSourceKeys = sSource.split('.');
            
            oSourceMap.sSourceName = aSourceKeys[0];
            aSourceKeys.shift();//Remove source and save remaining keys to determine data key at source

            if(aSourceKeys.length >= 0){
                oSourceMap.aDataPath = aSourceKeys;
            }            
        }
        else{
            oSourceMap.sSourceName = sSource;
        }

        return oSourceMap;        
    }

    //Used to retreive a single piece of data from a source. 

    //sFullSource resolves to a piece of data/object. sDataSet can be used to retreive multiple properties from that data/object.

    /*
        sFullSource = "propertyInfo.address"
        aDataSet= {
            "city": "city",
            "state": "state",    
            "street": "street1"
            "ownerName":"structure1.color"
        };

        would return a populated data response composed of:
        {
            "city": "propertyInfo.address.city",
            "state": "propertyInfo.address.state",    
            "street": "propertyInfo.address.street1"
            "ownerName":"propertyInfo.address.structure1.color"
        };

    */   

}
